from math import ceil
from typing import List

from pyxis.astro.coordinates import GCRFstate
from pyxis.math.constants import SECONDS_IN_DAY
from pyxis.math.linalg import Vector3D
from pyxis.time import Epoch


class RK4:

    #: Largest step to be taken by the integrator
    MAX_STEP = 300

    def __init__(self, state: GCRFstate) -> None:
        """class used to propagate a satellite state

        :param state: ECI state of the satellite to be propagated
        :type state: GCRFstate
        """
        #: the current state of the propagator
        self.state: GCRFstate = state.copy()

        #: integration step to be taken when the propagator is advanced
        self.step_size: float = RK4.MAX_STEP

    def step(self) -> None:
        """advance the propagator state by the stored time step"""
        h = self.step_size

        epoch_0: Epoch = self.state.epoch.copy()

        y: List[Vector3D] = self.state.vector_list()

        k1: List[Vector3D] = self.state.derivative()

        dsecs: float = h / 2
        ddays: float = dsecs / SECONDS_IN_DAY
        epoch_1 = epoch_0.plus_days(ddays)
        y1: GCRFstate = GCRFstate(epoch_1, y[0].plus(k1[0].scaled(dsecs)), y[1].plus(k1[1].scaled(dsecs)))
        k2: List[Vector3D] = y1.derivative()

        y2: GCRFstate = GCRFstate(epoch_1, y[0].plus(k2[0].scaled(dsecs)), y[1].plus(k2[1].scaled(dsecs)))
        k3: List[Vector3D] = y2.derivative()

        epoch_2 = epoch_1.plus_days(ddays)
        y3: GCRFstate = GCRFstate(epoch_2, y[0].plus(k3[0].scaled(h)), y[1].plus(k3[1].scaled(h)))
        k4: List[Vector3D] = y3.derivative()

        coeff: float = 1 / 6
        dv: Vector3D = k1[0].plus(k2[0].scaled(2).plus(k3[0].scaled(2).plus(k4[0]))).scaled(coeff)
        da: Vector3D = k1[1].plus(k2[1].scaled(2).plus(k3[1].scaled(2).plus(k4[1]))).scaled(coeff)

        self.state = GCRFstate(
            epoch_2,
            self.state.position.plus(dv.scaled(h)),
            self.state.velocity.plus(da.scaled(h)),
        )

    def step_to_epoch(self, epoch: Epoch) -> None:
        """advance the propagator state to the argument epoch

        :param epoch: time of state to be calculated
        :type epoch: Epoch
        """

        # Calculate time delta in seconds
        dt = (epoch.value - self.state.epoch.value) * SECONDS_IN_DAY

        # Determine number of steps required to meet new epoch while staying below the maximum step
        num_steps = ceil(abs(dt / self.MAX_STEP))

        # Store current step size
        old_step = self.step_size

        # Temporarily set step size to calculated dt
        if num_steps > 0:
            self.step_size = dt / num_steps

        # Step until desired epoch is achieved
        step_n = 0
        while step_n < num_steps:
            self.step()
            step_n += 1

        # Reset step size
        self.step_size = old_step



from math import asin, atan2, cos, pi, radians, sin, sqrt
from typing import List

from pyxis.astro.bodies.celestial import Earth, Moon, Sun
from pyxis.math.functions import Conversions, sign
from pyxis.math.linalg import Matrix3D, Vector3D, Vector6D
from pyxis.time import Epoch


class HillState:
    def __init__(self, position: Vector3D, velocity: Vector3D) -> None:
        """used to perform relative motion operations for two spacecraft represented in the Hill frame

        :param position: relative position of the two spacecraft in km
        :type position: Vector3D
        :param velocity: relative velocity of the two spacecraft in km/s
        :type velocity: Vector3D
        """
        #: relative position of the two spacecraft in km
        self.position: Vector3D = position.copy()

        #: relative velocity of the two spacecraft in km/s
        self.velocity: Vector3D = velocity.copy()

        #: state vector whose elements are equal to that of the position and velocity unpacked
        self.vector = Vector6D.from_position_and_velocity(self.position, self.velocity)

    @classmethod
    def from_state_vector(cls, state: Vector6D) -> "HillState":
        """create a HillState from a 6-D state vector

        :param state: state vector with components of the position and velocity in km and km/s respectively
        :type state: Vector6D
        :return: the 6-D vector represented as a relative state object
        :rtype: HillState
        """
        return cls(Vector3D(state.x, state.y, state.z), Vector3D(state.vx, state.vy, state.vz))

    @classmethod
    def from_gcrf(cls, state: "GCRFstate", origin: "GCRFstate") -> "HillState":
        """create a relative state from two inertial states

        :param state: inertial state of the chase vehicle
        :type state: GCRFstate
        :param origin: inertial state that will act as the origin for the frame
        :type origin: GCRFstate
        :return: relative state of the two inertial states
        :rtype: HillState
        """
        magrtgt: float = origin.position.magnitude()
        magrint: float = state.position.magnitude()
        rot_eci_rsw: Matrix3D = HillState.frame_matrix(origin)
        vtgtrsw: Vector3D = rot_eci_rsw.multiply_vector(origin.velocity)
        rintrsw: Vector3D = rot_eci_rsw.multiply_vector(state.position)
        vintrsw: Vector3D = rot_eci_rsw.multiply_vector(state.velocity)

        sinphiint: float = rintrsw.z / magrint
        phiint: float = asin(sinphiint)
        cosphiint: float = cos(phiint)
        lambdaint: float = atan2(rintrsw.y, rintrsw.x)
        sinlambdaint: float = sin(lambdaint)
        coslambdaint: float = cos(lambdaint)
        lambdadottgt: float = vtgtrsw.y / magrtgt

        rhill: Vector3D = Vector3D(magrint - magrtgt, lambdaint * magrtgt, phiint * magrtgt)

        rot_rsw_sez: Matrix3D = Matrix3D(
            Vector3D(sinphiint * coslambdaint, sinphiint * sinlambdaint, -cosphiint),
            Vector3D(-sinlambdaint, coslambdaint, 0),
            Vector3D(cosphiint * coslambdaint, cosphiint * sinlambdaint, sinphiint),
        )

        vintsez: Vector3D = rot_rsw_sez.multiply_vector(vintrsw)
        phidotint: float = -vintsez.x / magrint
        lambdadotint: float = vintsez.y / (magrint * cosphiint)

        vhill: Vector3D = Vector3D(
            vintsez.z - vtgtrsw.x,
            magrtgt * (lambdadotint - lambdadottgt),
            magrtgt * phidotint,
        )

        return cls(rhill, vhill)

    @staticmethod
    def frame_matrix(origin: "GCRFstate") -> Matrix3D:
        """create a radial, in-track, cross-track axes matrix

        :param origin: inertial state that acts as the origin for the RIC frame
        :type origin: GCRFstate
        :return: matrix with rows of radial, in-track, and cross-track
        :rtype: Matrix3D
        """
        r: Vector3D = origin.position.normalized()
        c: Vector3D = origin.position.cross(origin.velocity).normalized()
        i: Vector3D = c.cross(r)
        return Matrix3D(r, i, c)

    def copy(self) -> "HillState":
        """create a replica of the calling state

        :return: state with elements equal to that of the calling state
        :rtype: HillState
        """
        return HillState(self.position, self.velocity)

    def to_gcrf(self, origin: "GCRFstate") -> "GCRFstate":
        """create an inertial state for the calling state

        :param origin: inertial state that acts as the origin for the relative state
        :type origin: GCRFstate
        :return: inertial state of the relative spacecraft
        :rtype: GCRFstate
        """
        magrtgt: float = origin.position.magnitude()
        magrint: float = magrtgt + self.position.x
        rot_eci_rsw: Matrix3D = HillState.frame_matrix(origin)
        vtgtrsw: Vector3D = rot_eci_rsw.multiply_vector(origin.velocity)

        lambdadottgt: float = vtgtrsw.y / magrtgt
        lambdaint: float = self.position.y / magrtgt
        phiint: float = self.position.z / magrtgt
        sinphiint: float = sin(phiint)
        cosphiint: float = cos(phiint)
        sinlambdaint: float = sin(lambdaint)
        coslambdaint: float = cos(lambdaint)

        rot_rsw_sez: Matrix3D = Matrix3D(
            Vector3D(sinphiint * coslambdaint, sinphiint * sinlambdaint, -cosphiint),
            Vector3D(-sinlambdaint, coslambdaint, 0),
            Vector3D(cosphiint * coslambdaint, cosphiint * sinlambdaint, sinphiint),
        )

        rdotint: float = self.velocity.x + vtgtrsw.x
        lambdadotint: float = self.velocity.y / magrtgt + lambdadottgt
        phidotint: float = self.velocity.z / magrtgt
        vintsez: Vector3D = Vector3D(-magrint * phidotint, magrint * lambdadotint * cosphiint, rdotint)
        vintrsw: Vector3D = rot_rsw_sez.transpose().multiply_vector(vintsez)
        vinteci: Vector3D = rot_eci_rsw.transpose().multiply_vector(vintrsw)

        rintrsw: Vector3D = Vector3D(
            cosphiint * magrint * coslambdaint,
            cosphiint * magrint * sinlambdaint,
            sinphiint * magrint,
        )

        rinteci: Vector3D = rot_eci_rsw.transpose().multiply_vector(rintrsw)

        return GCRFstate(origin.epoch, rinteci, vinteci)


class GCRFstate:
    def __init__(self, epoch: Epoch, position: Vector3D, velocity: Vector3D) -> None:
        """used to perform operations in the ECI frame

        :param epoch: valid time of the state
        :type epoch: Epoch
        :param position: position of the inertial state in km
        :type position: Vector3D
        :param velocity: velocity of the inertial state in km/s
        :type velocity: Vector3D
        """
        #: valid time of the state
        self.epoch: Epoch = epoch.copy()

        #: position of the inertial state in km
        self.position: Vector3D = position.copy()

        #: velocity of the inertial state in km/s
        self.velocity: Vector3D = velocity.copy()

    @classmethod
    def from_hill(cls, origin: "GCRFstate", state: HillState) -> "GCRFstate":
        """create an inertial state from a relative state

        :param origin: inertial state that acts as the origin for the relative system
        :type origin: GCRFstate
        :param state: relative state of the spacecraft
        :type state: HillState
        :return: inertial state of the spacecraft
        :rtype: GCRFstate
        """
        return state.to_gcrf(origin)

    def copy(self) -> "GCRFstate":
        """create a replica of the calling state

        :return: state with attributes equal to the calling state
        :rtype: GCRFstate
        """
        return GCRFstate(self.epoch, self.position, self.velocity)

    def vector_list(self) -> List[Vector3D]:
        """create a list with elements of 0 == position and 1 == velocity

        :return: list of position and velocity
        :rtype: List[Vector3D]
        """
        return [self.position.copy(), self.velocity.copy()]

    def acceleration_from_earth(self) -> Vector3D:
        """calculate the acceleration on the state due to earth's gravity

        :return: vector representing the acceleration from earth
        :rtype: Vector3D
        """
        r_mag: float = self.position.magnitude()
        return self.position.scaled(-Earth.MU / (r_mag * r_mag * r_mag))

    def acceleration_from_moon(self) -> Vector3D:
        """calculate the acceleration on the state due to the moon

        :return: vector representing the acceleration from the moon
        :rtype: Vector3D
        """
        s: Vector3D = Moon.get_position(self.epoch)
        r: Vector3D = s.minus(self.position)
        r_mag: float = r.magnitude()
        s_mag: float = s.magnitude()
        vec_1: Vector3D = r.scaled(1 / (r_mag * r_mag * r_mag))
        vec_2: Vector3D = s.scaled(1 / (s_mag * s_mag * s_mag))
        return vec_1.minus(vec_2).scaled(Moon.MU)

    def acceleration_from_sun(self) -> Vector3D:
        """calculate the acceleration on the state due to the sun

        :return: vector representing the acceleration from the sun
        :rtype: Vector3D
        """
        s: Vector3D = Sun.get_position(self.epoch)
        r: Vector3D = s.minus(self.position)
        r_mag: float = r.magnitude()
        s_mag: float = s.magnitude()
        vec_1: Vector3D = r.scaled(1 / (r_mag * r_mag * r_mag))
        vec_2: Vector3D = s.scaled(1 / (s_mag * s_mag * s_mag))
        return vec_1.minus(vec_2).scaled(Sun.MU)

    def acceleration_from_srp(self) -> Vector3D:
        """calculate the acceleration on the state from solar radiation pressure

        :return: vector representing the acceleration from srp
        :rtype: Vector3D
        """
        sun_vec: Vector3D = self.sun_vector().normalized()
        return sun_vec.scaled(-Sun.P * 3.6e-5)

    def derivative(self) -> List[Vector3D]:
        """create a list with elements 0 == velocity and 1 == acceleration

        :return: list of velocity and acceleration
        :rtype: List[Vector3D]
        """
        net_0: Vector3D = Vector3D(0, 0, 0)
        net_1: Vector3D = net_0.plus(self.acceleration_from_moon())
        net_2: Vector3D = net_1.plus(self.acceleration_from_sun())
        net_3: Vector3D = net_2.plus(self.acceleration_from_srp())
        net_a: Vector3D = net_3.plus(self.acceleration_from_earth())
        return [self.velocity.copy(), net_a]

    def sun_vector(self) -> Vector3D:
        """create a vector pointing from the calling state to the sun

        :return: vector originating at the calling state and terminating at the sun
        :rtype: Vector3D
        """
        return Sun.get_position(self.epoch).minus(self.position)

    def moon_vector(self) -> Vector3D:
        """create a vector pointing from the calling state to the moon

        :return: vector originating at the calling state and terminating at the moon
        :rtype: Vector3D
        """
        return Moon.get_position(self.epoch).minus(self.position)

    def itrf_position(self) -> Vector3D:
        """create a vector that represents the state's position in the earth-fixed frame

        :return: the position of the state in the itrf frame
        :rtype: Vector3D
        """
        mod: Vector3D = Precession.matrix(self.epoch).multiply_vector(self.position)
        tod: Vector3D = Nutation.matrix(self.epoch).multiply_vector(mod)
        return Rotation.matrix(self.epoch).multiply_vector(tod)


class Rotation:
    """class used to transform between itrf and tod

    :return: defined in static methods
    :rtype: defined in static methods
    """

    @staticmethod
    def matrix(epoch: Epoch) -> Matrix3D:
        """creates a matrix that can be used to transform an itrf position to tod and vice versa

        :param epoch: valid time of the state
        :type epoch: Epoch
        :return: transformation matrix
        :rtype: Matrix3D
        """
        d: float = epoch.julian_value() - Epoch.J2000_JULIAN_DATE
        arg1: float = radians(125.0 - 0.05295 * d)
        arg2: float = radians(200.9 + 1.97129 * d)
        a: float = radians(-0.0048)
        b: float = radians(0.0004)
        dpsi: float = a * sin(arg1) - b * sin(arg2)
        eps: float = Earth.obliquity_of_ecliptic_at_epoch(epoch)
        gmst: float = epoch.greenwich_hour_angle()
        gast: float = dpsi * cos(eps) + gmst
        return Vector3D.rotation_matrix(Vector3D(0, 0, 1), -gast)


class Precession:
    """class used to transform between tod and mod

    :return: defined in static methods
    :rtype: defined in static methods
    """

    @staticmethod
    def matrix(epoch: Epoch) -> Matrix3D:
        """creates a matrix that can be used to transform a tod position to mod and vice versa

        :param epoch: valid time of the state
        :type epoch: Epoch
        :return: transformation matrix
        :rtype: Matrix3D
        """
        t: float = epoch.julian_centuries_past_j2000()
        a: float = Conversions.dms_to_radians(0, 0, 2306.2181)
        b: float = Conversions.dms_to_radians(0, 0, 0.30188)
        c: float = Conversions.dms_to_radians(0, 0, 0.017998)
        x: float = a * t + b * t * t + c * t * t * t

        a = Conversions.dms_to_radians(0, 0, 2004.3109)
        b = Conversions.dms_to_radians(0, 0, 0.42665)
        c = Conversions.dms_to_radians(0, 0, 0.041833)
        y: float = a * t - b * t * t - c * t * t * t

        a = Conversions.dms_to_radians(0, 0, 0.7928)
        b = Conversions.dms_to_radians(0, 0, 0.000205)
        z: float = x + a * t * t + b * t * t * t

        sz = sin(z)
        sy = sin(y)
        sx = sin(x)
        cz = cos(z)
        cy = cos(y)
        cx = cos(x)

        return Matrix3D(
            Vector3D(-sz * sx + cz * cy * cx, -sz * cx - cz * cy * sx, -cz * sy),
            Vector3D(cz * sx + sz * cy * cx, cz * cx - sz * cy * sx, -sz * sy),
            Vector3D(sy * cx, -sy * sx, cy),
        )


class Nutation:
    """class used to transform between mod and gcrf

    :return: defined in static methods
    :rtype: defined in static methods
    """

    @staticmethod
    def matrix(epoch: Epoch) -> Matrix3D:
        """creates a matrix that can be used to transform a mod position to gcrf and vice versa

        :param epoch: valid time of the state
        :type epoch: Epoch
        :return: transformation matrix
        :rtype: Matrix3D
        """
        d = epoch.julian_value() - Epoch.J2000_JULIAN_DATE
        arg1 = radians(125.0 - 0.05295 * d)
        arg2 = radians(200.9 + 1.97129 * d)
        a = radians(-0.0048)
        b = radians(0.0004)
        dpsi = a * sin(arg1) - b * sin(arg2)
        a = radians(0.0026)
        b = radians(0.0002)
        deps = a * cos(arg1) + b * cos(arg2)
        eps = Earth.obliquity_of_ecliptic_at_epoch(epoch)

        ce = cos(eps)
        se = sin(eps)

        return Matrix3D(
            Vector3D(1.0, -dpsi * ce, -dpsi * se),
            Vector3D(dpsi * ce, 1.0, -deps),
            Vector3D(dpsi * se, deps, 1.0),
        )


class ITRFstate:
    def __init__(self, epoch: Epoch, position: Vector3D, velocity: Vector3D) -> None:
        """used to represent states in the earth-fixed frame

        :param epoch: valid time of the state
        :type epoch: Epoch
        :param position: position of the state in km
        :type position: Vector3D
        :param velocity: velocity of the state in km/s
        :type velocity: Vector3D
        """
        #: valid time of the state
        self.epoch: Epoch = epoch.copy()

        #: position of the state in km
        self.position: Vector3D = position.copy()

        #: velocity of the state in km/s
        self.velocity: Vector3D = velocity.copy()

    def gcrf_position(self) -> Vector3D:
        """create a vector that represents the position of the calling state in the inertial frame

        :return: ECI position of the calling state
        :rtype: Vector3D
        """
        tod: Vector3D = Rotation.matrix(self.epoch).transpose().multiply_vector(self.position)
        mod: Vector3D = Nutation.matrix(self.epoch).transpose().multiply_vector(tod)
        return Precession.matrix(self.epoch).transpose().multiply_vector(mod)

    def lla_state(self) -> "LLAstate":
        """calculate the latitude, longitude, and altitude of the ecf position

        :return: geodetic lat, long, altitude of the ecf position
        :rtype: LLAstate
        """
        pos: Vector3D = self.position
        x: float = pos.x
        y: float = pos.y
        z: float = pos.z

        # Equation 2.77a
        a2: float = Earth.RADIUS * Earth.RADIUS
        f: float = Earth.FLATTENING
        e2: float = f * (2.0 - f)
        b2: float = a2 - e2 * a2
        eps2: float = a2 / b2 - 1.0
        rho: float = sqrt(x * x + y * y)

        # Equation 2.77b
        p: float = abs(z) / eps2
        s: float = rho * rho / (e2 * eps2)
        q: float = p * p - b2 + s

        # Equation 2.77c
        u: float = p / sqrt(q)
        v: float = b2 * u * u / q
        cap_p: float = 27.0 * v * s / q
        cap_q: float = (sqrt(cap_p + 1) + sqrt(cap_p)) ** 2.0 / 3.0

        # Equation 2.77d
        t: float = (1.0 + cap_q + 1.0 / cap_q) / 6.0
        c: float = sqrt(u * u - 1.0 + 2.0 * t)
        w: float = (c - u) / 2.0

        # Equation 2.77e
        base: float = sqrt(t * t + v) - u * w - t / 2.0 - 0.25
        arg: float = w + sqrt(base)
        d: float = sign(z) * sqrt(q) * arg

        # Equation 2.77f
        n: float = sqrt(a2) * sqrt(1.0 + eps2 * d * d / b2)
        lamb: float = asin((eps2 + 1.0) * (d / n))

        # Equation 2.77g
        h: float = rho * cos(lamb) + z * sin(lamb) - a2 / n
        phi: float = atan2(y, x)
        if phi < 0:
            phi += pi * 2.0

        return LLAstate(lamb, phi, h)


class LLAstate:
    def __init__(self, lat: float, longit: float, alt: float) -> None:
        """used to perform operations for a state in an oblate earth frame

        :param lat: geodetic latitude in radians
        :type lat: float
        :param long: geodetic longitude in radians
        :type long: float
        :param alt: altitude above the surface in km
        :type alt: float
        """
        #: geodetic latitude in radians
        self.latitude: float = lat

        #: geodetic longitude in radians
        self.longitude: float = longit

        #: altitude above the surface in km
        self.altitude: float = alt

    def copy(self) -> "LLAstate":
        """creates a duplicate of the calling state

        :return: state with properties that match that of the calling state
        :rtype: LLAstate
        """
        return LLAstate(self.latitude, self.longitude, self.altitude)

    def itrf_position(self) -> Vector3D:
        """creates a cartesian vector to represent the lat, long, and altitude

        :return: the earth-fixed cartesian position of the lat, long, and altitude
        :rtype: Vector3D
        """
        lat: float = self.latitude
        longitude: float = self.longitude
        alt: float = self.altitude

        f: float = Earth.FLATTENING
        e: float = sqrt(f * (2 - f))
        slat: float = sin(lat)
        clat: float = cos(lat)
        n: float = Earth.RADIUS / sqrt(1 - e * e * slat * slat)

        return Vector3D(
            (n + alt) * clat * cos(longitude), (n + alt) * clat * sin(longitude), (n * (1.0 - e * e) + alt) * slat
        )


class AzElRange:
    def __init__(self, az: float, el: float, r: float) -> None:
        """used to perform operations related to ground site measurements

        :param az: clock-wise angle from the north vector
        :type az: float
        :param el: angle measured from the horizon plane
        :type el: float
        :param r: distance to the observed object
        :type r: float
        """
        self.azimuth: float = az
        self.elevation: float = el
        self.range: float = r

    @classmethod
    def from_enz(cls, enz: Vector3D) -> "AzElRange":
        """calculate the azimuth, elevation, and range given a vector in the enz frame

        :param enz: vector in the east-north-zenith frame
        :type enz: Vector3D
        :return: topo-centric azimuth, elevation, and range
        :rtype: AzElRange
        """
        az: float = atan2(enz.x, enz.y)
        if az < 0:
            az += pi * 2.0
        el: float = atan2(enz.z, sqrt(enz.x * enz.x + enz.y * enz.y))
        return cls(az, el, enz.magnitude())



from math import acos, cos, sin, sqrt


class Vector6D:
    def __init__(self, x: float, y: float, z: float, vx: float, vy: float, vz: float) -> None:
        """class used to perform operations on a vector that has 6 components

        :param x: first component of vector
        :type x: float
        :param y: second component of vector
        :type y: float
        :param z: third component of vector
        :type z: float
        :param vx: fourth component of vector
        :type vx: float
        :param vy: fifth component of vector
        :type vy: float
        :param vz: sixth component of vector
        :type vz: float
        """
        #: first component of the vector
        self.x: float = x

        #: second component of the vector
        self.y: float = y

        #: third component of the vector
        self.z: float = z

        #: fourth component of the vector
        self.vx: float = vx

        #: fifth component of the vector
        self.vy: float = vy

        #: sixth component of the vector
        self.vz: float = vz

    @classmethod
    def from_position_and_velocity(cls, r: "Vector3D", v: "Vector3D") -> "Vector6D":
        """create a 6-dimension vecto from 3-D position and velocity

        :param r: 3-D position vector
        :type r: Vector3D
        :param v: 3-D velocity vector
        :type v: Vector3D
        :return: 6-D state vector
        :rtype: Vector6D
        """
        return cls(r.x, r.y, r.z, v.x, v.y, v.z)

    def __str__(self) -> str:
        """overrides the string method so stored values will be printed

        :return: string representation of the vector
        :rtype: str
        """
        return f"{self.x:.6f} {self.y:.6f} {self.z:.6f} {self.vx:.6f} {self.vy:.6f} {self.vz:.6f}"

    def dot(self, vec_to_dot: "Vector6D") -> float:
        """calculates the dot product of the two vectors

        :param vec_to_dot: the second vector to be used in the dot product
        :type vec_to_dot: Vector6D
        :return: sum of the element products
        :rtype: float
        """
        return (
            self.x * vec_to_dot.x
            + self.y * vec_to_dot.y
            + self.z * vec_to_dot.z
            + self.vx * vec_to_dot.vx
            + self.vy * vec_to_dot.vy
            + self.vz * vec_to_dot.vz
        )

    def copy(self) -> "Vector6D":
        """create a replica of the current vector

        :return: 6-D vector with components that match the calling vector
        :rtype: Vector6D
        """
        return Vector6D(self.x, self.y, self.z, self.vx, self.vy, self.vz)

    def plus(self, vec: "Vector6D") -> "Vector6D":
        """calculates the sum of the elements of the two vectors

        :param vec: second vector to be included in the sum
        :type vec: Vector6D
        :return: vector whose elements are sums of the calling and argument vector
        :rtype: Vector6D
        """
        return Vector6D(
            self.x + vec.x,
            self.y + vec.y,
            self.z + vec.z,
            self.vx + vec.vx,
            self.vy + vec.vy,
            self.vz + vec.vz,
        )

    def minus(self, vec: "Vector6D") -> "Vector6D":
        """calculates the difference of the elements of the two vectors

        :param vec: second vector to be included in the difference
        :type vec: Vector6D
        :return: vector whose elements are the difference of the calling and argument vector
        :rtype: Vector6D
        """
        return Vector6D(
            self.x - vec.x,
            self.y - vec.y,
            self.z - vec.z,
            self.vx - vec.vx,
            self.vy - vec.vy,
            self.vz - vec.vz,
        )


class Vector3D:
    def __init__(self, x: float, y: float, z: float) -> None:
        """class used to perform operations on a 3-dimension vector

        :param x: first component of the vector
        :type x: float
        :param y: second component of the vector
        :type y: float
        :param z: third component of the vector
        :type z: float
        """
        #: first component of the vector
        self.x: float = x

        #: second component of the vector
        self.y: float = y

        #: third component of the vector
        self.z: float = z

    def __str__(self) -> str:
        """overrides the string method so the elements can be viewed

        :return: string representation of the vector elements
        :rtype: str
        """
        return f"{self.x:.6f} {self.y:.6f} {self.z:.6f}"

    def copy(self) -> "Vector3D":
        """creates a replica of the vector

        :return: 3-D vector with elements equal to the calling vector
        :rtype: Vector3D
        """
        return Vector3D(self.x, self.y, self.z)

    def plus(self, vec_to_add: "Vector3D") -> "Vector3D":
        """creates a vector whose elements equal the sum of the calling and argument vector

        :param vec_to_add: vector to be included in the sum
        :type vec_to_add: Vector3D
        :return: vector with elements equal to the sum of the calling and argument vector
        :rtype: Vector3D
        """
        return Vector3D(self.x + vec_to_add.x, self.y + vec_to_add.y, self.z + vec_to_add.z)

    def minus(self, vec_to_subtract: "Vector3D") -> "Vector3D":
        """creates a vector whose elements equal the difference of the calling and argument vector

        :param vec_to_subtract: vector to be included in the difference
        :type vec_to_subtract: Vector3D
        :return: vector with elements equal to the difference of the calling and argument vector
        :rtype: Vector3D
        """
        return Vector3D(
            self.x - vec_to_subtract.x,
            self.y - vec_to_subtract.y,
            self.z - vec_to_subtract.z,
        )

    def dot(self, vec_to_dot: "Vector3D") -> float:
        """calculates the dot product of the two vectors

        :param vec_to_dot: the second vector to be used in the dot product
        :type vec_to_dot: Vector3D
        :return: sum of the element products
        :rtype: float
        """
        return self.x * vec_to_dot.x + self.y * vec_to_dot.y + self.z * vec_to_dot.z

    def cross(self, vec_to_cross: "Vector3D") -> "Vector3D":
        """creates a vector orthogonal to the calling and argument vector

        :param vec_to_cross: vector which will be used to complete the right-hand rule
        :type vec_to_cross: Vector3D
        :return: vector produced using the right-hand rule that is orthogonal to the original vectors
        :rtype: Vector3D
        """
        return Vector3D(
            self.y * vec_to_cross.z - self.z * vec_to_cross.y,
            self.z * vec_to_cross.x - self.x * vec_to_cross.z,
            self.x * vec_to_cross.y - self.y * vec_to_cross.x,
        )

    def magnitude(self) -> float:
        """calculates the length of the vector

        :return: square root of the sum of squares
        :rtype: float
        """
        return sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def scaled(self, scalar: float) -> "Vector3D":
        """creates a scaled copy of the calling vector

        :param scalar: value that will be multiplied against all elements
        :type scalar: float
        :return: vector equal to the calling vector scaled by the argument
        :rtype: Vector3D
        """
        return Vector3D(self.x * scalar, self.y * scalar, self.z * scalar)

    def normalized(self) -> "Vector3D":
        """creates a vector parallel to the calling vector that is of length 1

        :return: unit vector of the calling vector
        :rtype: Vector3D
        """
        return self.scaled(1 / self.magnitude())

    def angle(self, adj_vec: "Vector3D") -> float:
        """calculates the angle between two vectors

        :param adj_vec: second leg of the angle
        :type adj_vec: Vector3D
        :return: angle in radians between the two vectors
        :rtype: float
        """
        arg = self.dot(adj_vec) / (self.magnitude() * adj_vec.magnitude())
        if arg > 1:
            arg = 1
        elif arg < -1:
            arg = -1
        return acos(arg)

    @staticmethod
    def rotation_matrix(axis: "Vector3D", theta: float) -> "Matrix3D":
        """calculates the rotation matrix required for a rotation

        :param axis: axis that will be used to rotate a vector around
        :type axis: Vector3D
        :param theta: angle in radians the vector will be rotated
        :type theta: float
        :return: transition matrix that will rotate a vector
        :rtype: Matrix3D
        """
        unit_ax: Vector3D = axis.normalized()
        ux: float = unit_ax.x
        uy: float = unit_ax.y
        uz: float = unit_ax.z
        c: float = cos(theta)
        s: float = sin(theta)
        cdiff: float = 1 - c

        x1: float = c + ux * ux * cdiff
        y1: float = ux * uy * cdiff - uz * s
        z1: float = ux * uz * cdiff + uy * s
        x2: float = uy * ux * cdiff + uz * s
        y2: float = c + uy * uy * cdiff
        z2: float = uy * uz * cdiff - ux * s
        x3: float = uz * ux * cdiff - uy * s
        y3: float = uz * uy * cdiff + ux * s
        z3: float = c + uz * uz * cdiff

        r1: Vector3D = Vector3D(x1, y1, z1)
        r2: Vector3D = Vector3D(x2, y2, z2)
        r3: Vector3D = Vector3D(x3, y3, z3)

        return Matrix3D(r1, r2, r3)

    def rotation_about_axis(self, axis: "Vector3D", theta: float) -> "Vector3D":
        """create a vector that has been rotated

        :param axis: axis of rotation
        :type axis: Vector3D
        :param theta: angle in radians of rotation
        :type theta: float
        :return: vector that has been rotated
        :rtype: Vector3D
        """
        return Vector3D.rotation_matrix(axis, theta).multiply_vector(self.copy())


class Matrix3D:
    def __init__(self, row1: Vector3D, row2: Vector3D, row3: Vector3D) -> None:
        """used to perform operations on a 3x3 matrix

        :param row1: first row of matrix
        :type row1: Vector3D
        :param row2: second row of matrix
        :type row2: Vector3D
        :param row3: third row of matrix
        :type row3: Vector3D
        """
        #: first row of matrix
        self.row1: Vector3D = row1.copy()

        #: second row of matrix
        self.row2: Vector3D = row2.copy()

        #: third row of matrix
        self.row3: Vector3D = row3.copy()

    def diagonal(self) -> Vector3D:
        """creates a vector with components equal to the diagonal of the matrix

        :return: vector equal to (xx, yy, zz)
        :rtype: Vector3D
        """
        return Vector3D(self.row1.x, self.row2.y, self.row3.z)

    def column_1(self) -> Vector3D:
        """creates a vector with elements equal to the first column of the matrix

        :return: vector with elements equal to the first column of the matrix
        :rtype: Vector3D
        """
        return Vector3D(self.row1.x, self.row2.x, self.row3.x)

    def column_2(self) -> Vector3D:
        """creates a vector with elements equal to the second column of the matrix

        :return: vector with elements equal to the second column of the matrix
        :rtype: Vector3D
        """
        return Vector3D(self.row1.y, self.row2.y, self.row3.y)

    def column_3(self) -> Vector3D:
        """creates a vector with elements equal to the third column of the matrix

        :return: vector with elements equal to the third column of the matrix
        :rtype: Vector3D
        """
        return Vector3D(self.row1.z, self.row2.z, self.row3.z)

    def multiply_vector(self, vec: Vector3D) -> Vector3D:
        """performs matrix multiplication of the calling matrix and the argument vector

        :param vec: vector to be used in the multiplication
        :type vec: Vector3D
        :return: product of the matrix multiplication
        :rtype: Vector3D
        """
        return Vector3D(self.row1.dot(vec), self.row2.dot(vec), self.row3.dot(vec))

    def scaled(self, scalar: float) -> "Matrix3D":
        """creates a matrix whose elements have been scaled by the argument

        :param scalar: multiple to be used in the scale
        :type scalar: float
        :return: matrix with elements equal to the original scaled by the multiple
        :rtype: Matrix3D
        """
        return Matrix3D(self.row1.scaled(scalar), self.row2.scaled(scalar), self.row3.scaled(scalar))

    def transpose(self) -> "Matrix3D":
        """creates a matrix whose rows are equal to the columns of the calling matrix

        :return: transposed matrix of the calling matrix
        :rtype: Matrix3D
        """
        return Matrix3D(
            Vector3D(self.row1.x, self.row2.x, self.row3.x),
            Vector3D(self.row1.y, self.row2.y, self.row3.y),
            Vector3D(self.row1.z, self.row2.z, self.row3.z),
        )

    def plus(self, mat: "Matrix3D") -> "Matrix3D":
        """creates a matrix whose elements are equal to the sum of the calling and argument matrix

        :param mat: matrix to be used in the sum
        :type mat: Matrix3D
        :return: sum matrix
        :rtype: Matrix3D
        """
        return Matrix3D(self.row1.plus(mat.row1), self.row2.plus(mat.row2), self.row3.plus(mat.row3))

    def determinant(self) -> float:
        """calculate the determinant of the matrix

        :return: determinant
        :rtype: float
        """
        return (
            self.row1.x * self.row2.y * self.row3.z
            + self.row1.y * self.row2.z * self.row3.x
            + self.row1.z * self.row2.x * self.row3.y
            - self.row1.z * self.row2.y * self.row3.x
            - self.row1.y * self.row2.x * self.row3.z
            - self.row1.x * self.row2.z * self.row3.y
        )

    def cofactor(self) -> "Matrix3D":
        """create a matrix that is the cofactor of the calling matrix

        :return: cofactor matrix
        :rtype: Matrix3D
        """
        return Matrix3D(
            Vector3D(
                self.row2.y * self.row3.z - self.row2.z * self.row3.y,
                -(self.row2.x * self.row3.z - self.row3.x * self.row2.z),
                self.row2.x * self.row3.y - self.row3.x * self.row2.y,
            ),
            Vector3D(
                -(self.row1.y * self.row3.z - self.row3.y * self.row1.z),
                self.row1.x * self.row3.z - self.row1.z * self.row3.x,
                -(self.row1.x * self.row3.y - self.row1.y * self.row3.x),
            ),
            Vector3D(
                self.row1.y * self.row2.z - self.row1.z * self.row2.y,
                -(self.row1.x * self.row2.z - self.row2.x * self.row1.z),
                self.row1.x * self.row2.y - self.row1.y * self.row2.x,
            ),
        )

    def adjugate(self) -> "Matrix3D":
        """create a matrix that is the adjugate of the calling matrix

        :return: adjugate matrix
        :rtype: Matrix3D
        """
        return self.cofactor().transpose()

    def inverse(self) -> "Matrix3D":
        """create a matrix that is the inverse of the calling matrix

        :return: inverse matrix
        :rtype: Matrix3D
        """
        return self.adjugate().scaled(1 / self.determinant())

    def multiply_matrix3by6(self, mat: "Matrix3by6") -> "Matrix3by6":
        """create a matrix that is the product of the calling 3x3 and an argument 3x6 matrix

        :param mat: 3x6 matrix to be used in the product
        :type mat: Matrix3by6
        :return: product matrix
        :rtype: Matrix3by6
        """
        return Matrix3by6(
            Vector6D(
                self.row1.dot(mat.column_1()),
                self.row1.dot(mat.column_2()),
                self.row1.dot(mat.column_3()),
                self.row1.dot(mat.column_4()),
                self.row1.dot(mat.column_5()),
                self.row1.dot(mat.column_6()),
            ),
            Vector6D(
                self.row2.dot(mat.column_1()),
                self.row2.dot(mat.column_2()),
                self.row2.dot(mat.column_3()),
                self.row2.dot(mat.column_4()),
                self.row2.dot(mat.column_5()),
                self.row2.dot(mat.column_6()),
            ),
            Vector6D(
                self.row3.dot(mat.column_1()),
                self.row3.dot(mat.column_2()),
                self.row3.dot(mat.column_3()),
                self.row3.dot(mat.column_4()),
                self.row3.dot(mat.column_5()),
                self.row3.dot(mat.column_6()),
            ),
        )


class Matrix3by6:
    def __init__(self, row1: Vector6D, row2: Vector6D, row3: Vector6D) -> None:
        """used to perform operations for a 3x6 matrix

        :param row1: first row of the matrix
        :type row1: Vector6D
        :param row2: second row of the matrix
        :type row2: Vector6D
        :param row3: third row of the matrix
        :type row3: Vector6D
        """
        self.row1: Vector6D = row1.copy()
        self.row2: Vector6D = row2.copy()
        self.row3: Vector6D = row3.copy()

    def column_1(self) -> Vector3D:
        """create a vector whose elements are equal to the first column of the calling matrix

        :return: first column
        :rtype: Vector3D
        """
        return Vector3D(self.row1.x, self.row2.x, self.row3.x)

    def column_2(self) -> Vector3D:
        """create a vector whose elements are equal to the second column of the calling matrix

        :return: second column
        :rtype: Vector3D
        """
        return Vector3D(self.row1.y, self.row2.y, self.row3.y)

    def column_3(self) -> Vector3D:
        """create a vector whose elements are equal to the third column of the calling matrix

        :return: third column
        :rtype: Vector3D
        """
        return Vector3D(self.row1.z, self.row2.z, self.row3.z)

    def column_4(self) -> Vector3D:
        """create a vector whose elements are equal to the fourth column of the calling matrix

        :return: fourth column
        :rtype: Vector3D
        """
        return Vector3D(self.row1.vx, self.row2.vx, self.row3.vx)

    def column_5(self) -> Vector3D:
        """create a vector whose elements are equal to the fifth column of the calling matrix

        :return: fifth column
        :rtype: Vector3D
        """
        return Vector3D(self.row1.vy, self.row2.vy, self.row3.vy)

    def column_6(self) -> Vector3D:
        """create a vector whose elements are equal to the sixth column of the calling matrix

        :return: sixth column
        :rtype: Vector3D
        """
        return Vector3D(self.row1.vz, self.row2.vz, self.row3.vz)

    def multiply_vector(self, vec: Vector6D) -> Vector3D:
        """create a vector that is the product of the calling matrix and the argument vector

        :param vec: vector to be used in the product
        :type vec: Vector6D
        :return: product vector
        :rtype: Vector3D
        """
        return Vector3D(self.row1.dot(vec), self.row2.dot(vec), self.row3.dot(vec))

    def transpose(self) -> "Matrix6by3":
        """create a matrix whose rows are equal to the columns of the calling matrix

        :return: the transpose of the calling matrix
        :rtype: Matrix6by3
        """
        return Matrix6by3(
            self.column_1(),
            self.column_2(),
            self.column_3(),
            self.column_4(),
            self.column_5(),
            self.column_6(),
        )

    def multiply_matrix_6by3(self, mat: "Matrix6by3") -> Matrix3D:
        """create a matrix equal to the product of the calling matrix and the argument matrix

        :param mat: matrix to be used in the product
        :type mat: Matrix6by3
        :return: product matrix
        :rtype: Matrix3D
        """
        return Matrix3D(
            Vector3D(
                self.row1.dot(mat.column_1()),
                self.row1.dot(mat.column_2()),
                self.row1.dot(mat.column_3()),
            ),
            Vector3D(
                self.row2.dot(mat.column_1()),
                self.row2.dot(mat.column_2()),
                self.row2.dot(mat.column_3()),
            ),
            Vector3D(
                self.row3.dot(mat.column_1()),
                self.row3.dot(mat.column_2()),
                self.row3.dot(mat.column_3()),
            ),
        )


class Matrix6by3:
    def __init__(
        self,
        r1: Vector3D,
        r2: Vector3D,
        r3: Vector3D,
        r4: Vector3D,
        r5: Vector3D,
        r6: Vector3D,
    ) -> None:
        """used to perform operations for a 6x3 matrix

        :param r1: first row of the matrix
        :type r1: Vector3D
        :param r2: second row of the matrix
        :type r2: Vector3D
        :param r3: third row of the matrix
        :type r3: Vector3D
        :param r4: fourth row of the matrix
        :type r4: Vector3D
        :param r5: fifth row of the matrix
        :type r5: Vector3D
        :param r6: sixth row of the matrix
        :type r6: Vector3D
        """
        #: first row of the matrix
        self.row1: Vector3D = r1.copy()

        #: second row of the matrix
        self.row2: Vector3D = r2.copy()

        #: third row of the matrix
        self.row3: Vector3D = r3.copy()

        #: fourth row of the matrix
        self.row4: Vector3D = r4.copy()

        #: fifth row of the matrix
        self.row5: Vector3D = r5.copy()

        #: sixth row of the matrix
        self.row6: Vector3D = r6.copy()

    def column_1(self) -> Vector6D:
        """create a vector whose elements equal the first column of the matrix

        :return: first column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(self.row1.x, self.row2.x, self.row3.x, self.row4.x, self.row5.x, self.row6.x)

    def column_2(self) -> Vector6D:
        """create a vector whose elements equal the second column of the matrix

        :return: second column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(self.row1.y, self.row2.y, self.row3.y, self.row4.y, self.row5.y, self.row6.y)

    def column_3(self) -> Vector6D:
        """create a vector whose elements equal the third column of the matrix

        :return: third column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(self.row1.z, self.row2.z, self.row3.z, self.row4.z, self.row5.z, self.row6.z)

    def transpose(self) -> Matrix3by6:
        """create a matrix whose rows equal the columns of the calling matrix

        :return: transpose of the calling matrix
        :rtype: Matrix3by6
        """
        return Matrix3by6(self.column_1(), self.column_2(), self.column_3())

    def multiply(self, mat: Matrix3D) -> "Matrix6by3":
        """create a matrix that is the product of the calling and argument matrices

        :param mat: matrix to be used in the product
        :type mat: Matrix3D
        :return: product matrix
        :rtype: Matrix6by3
        """
        return Matrix6by3(
            Vector3D(
                self.row1.dot(mat.column_1()),
                self.row1.dot(mat.column_2()),
                self.row1.dot(mat.column_3()),
            ),
            Vector3D(
                self.row2.dot(mat.column_1()),
                self.row2.dot(mat.column_2()),
                self.row2.dot(mat.column_3()),
            ),
            Vector3D(
                self.row3.dot(mat.column_1()),
                self.row3.dot(mat.column_2()),
                self.row3.dot(mat.column_3()),
            ),
            Vector3D(
                self.row4.dot(mat.column_1()),
                self.row4.dot(mat.column_2()),
                self.row4.dot(mat.column_3()),
            ),
            Vector3D(
                self.row5.dot(mat.column_1()),
                self.row5.dot(mat.column_2()),
                self.row5.dot(mat.column_3()),
            ),
            Vector3D(
                self.row6.dot(mat.column_1()),
                self.row6.dot(mat.column_2()),
                self.row6.dot(mat.column_3()),
            ),
        )

    def multiply_vector(self, vec: "Vector3D") -> Vector6D:
        """create a vector equal to the product of the calling matrix and the argument vector

        :param vec: vector to be used in the product
        :type vec: Vector3D
        :return: product vector
        :rtype: Vector6D
        """
        return Vector6D(
            self.row1.dot(vec),
            self.row2.dot(vec),
            self.row3.dot(vec),
            self.row4.dot(vec),
            self.row5.dot(vec),
            self.row6.dot(vec),
        )

    def multiply_matrix3by6(self, mat: "Matrix3by6") -> "Matrix6D":
        """create a matrix equal to the product of the calling and the argument matrices

        :param mat: matrix to be used in the product
        :type mat: Matrix3by6
        :return: product matrix
        :rtype: Matrix6D
        """
        return Matrix6D(
            Vector6D(
                self.row1.dot(mat.column_1()),
                self.row1.dot(mat.column_2()),
                self.row1.dot(mat.column_3()),
                self.row1.dot(mat.column_4()),
                self.row1.dot(mat.column_5()),
                self.row1.dot(mat.column_6()),
            ),
            Vector6D(
                self.row2.dot(mat.column_1()),
                self.row2.dot(mat.column_2()),
                self.row2.dot(mat.column_3()),
                self.row2.dot(mat.column_4()),
                self.row2.dot(mat.column_5()),
                self.row2.dot(mat.column_6()),
            ),
            Vector6D(
                self.row3.dot(mat.column_1()),
                self.row3.dot(mat.column_2()),
                self.row3.dot(mat.column_3()),
                self.row3.dot(mat.column_4()),
                self.row3.dot(mat.column_5()),
                self.row3.dot(mat.column_6()),
            ),
            Vector6D(
                self.row4.dot(mat.column_1()),
                self.row4.dot(mat.column_2()),
                self.row4.dot(mat.column_3()),
                self.row4.dot(mat.column_4()),
                self.row4.dot(mat.column_5()),
                self.row4.dot(mat.column_6()),
            ),
            Vector6D(
                self.row5.dot(mat.column_1()),
                self.row5.dot(mat.column_2()),
                self.row5.dot(mat.column_3()),
                self.row5.dot(mat.column_4()),
                self.row5.dot(mat.column_5()),
                self.row5.dot(mat.column_6()),
            ),
            Vector6D(
                self.row6.dot(mat.column_1()),
                self.row6.dot(mat.column_2()),
                self.row6.dot(mat.column_3()),
                self.row6.dot(mat.column_4()),
                self.row6.dot(mat.column_5()),
                self.row6.dot(mat.column_6()),
            ),
        )


class Matrix6D:
    def __init__(
        self,
        r1: Vector6D,
        r2: Vector6D,
        r3: Vector6D,
        r4: Vector6D,
        r5: Vector6D,
        r6: Vector6D,
    ) -> None:
        """used to perform operations for a 6x6 matrix

        :param r1: first row of the matrix
        :type r1: Vector6D
        :param r2: second row of the matrix
        :type r2: Vector6D
        :param r3: third row of the matrix
        :type r3: Vector6D
        :param r4: fourth row of the matrix
        :type r4: Vector6D
        :param r5: fifth row of the matrix
        :type r5: Vector6D
        :param r6: sixth row of the matrix
        :type r6: Vector6D
        """
        #: first row of the matrix
        self.row1: Vector6D = r1.copy()

        #: second row of the matrix
        self.row2: Vector6D = r2.copy()

        #: third row of the matrix
        self.row3: Vector6D = r3.copy()

        #: fourth row of the matrix
        self.row4: Vector6D = r4.copy()

        #: fifth row of the matrix
        self.row5: Vector6D = r5.copy()

        #: sixth row of the matrix
        self.row6: Vector6D = r6.copy()

    @classmethod
    def identity(cls) -> "Matrix6D":
        """create a matrix with a diagonal with elements of 1 and off-diagonal elements of 0

        :return: the identity matrix
        :rtype: Matrix6D
        """
        return cls(
            Vector6D(1, 0, 0, 0, 0, 0),
            Vector6D(0, 1, 0, 0, 0, 0),
            Vector6D(0, 0, 1, 0, 0, 0),
            Vector6D(0, 0, 0, 1, 0, 0),
            Vector6D(0, 0, 0, 0, 1, 0),
            Vector6D(0, 0, 0, 0, 0, 1),
        )

    def diagonal(self) -> Vector6D:
        """create a vector whose elements are equal to the diagonal of the matrix

        :return: vector whose elements equal the diagonal
        :rtype: Vector6D
        """
        return Vector6D(
            self.row1.x,
            self.row2.y,
            self.row3.z,
            self.row4.vx,
            self.row5.vy,
            self.row6.vz,
        )

    def multiply_vector(self, vec: Vector6D) -> Vector6D:
        """create a vector equal to the product of the calling matrix and the argument vector

        :param vec: vector to be used in the product
        :type vec: Vector6D
        :return: product vector
        :rtype: Vector6D
        """
        return Vector6D(
            self.row1.dot(vec),
            self.row2.dot(vec),
            self.row3.dot(vec),
            self.row4.dot(vec),
            self.row5.dot(vec),
            self.row6.dot(vec),
        )

    def column_1(self) -> Vector6D:
        """create a vector whose elements equal the first column of the matrix

        :return: first column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(self.row1.x, self.row2.x, self.row3.x, self.row4.x, self.row5.x, self.row6.x)

    def column_2(self) -> Vector6D:
        """create a vector whose elements equal the second column of the matrix

        :return: second column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(self.row1.y, self.row2.y, self.row3.y, self.row4.y, self.row5.y, self.row6.y)

    def column_3(self) -> Vector6D:
        """create a vector whose elements equal the third column of the matrix

        :return: third column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(self.row1.z, self.row2.z, self.row3.z, self.row4.z, self.row5.z, self.row6.z)

    def column_4(self) -> Vector6D:
        """create a vector whose elements equal the fourth column of the matrix

        :return: fourth column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(
            self.row1.vx,
            self.row2.vx,
            self.row3.vx,
            self.row4.vx,
            self.row5.vx,
            self.row6.vx,
        )

    def column_5(self) -> Vector6D:
        """create a vector whose elements equal the fifth column of the matrix

        :return: fifth column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(
            self.row1.vy,
            self.row2.vy,
            self.row3.vy,
            self.row4.vy,
            self.row5.vy,
            self.row6.vy,
        )

    def column_6(self) -> Vector6D:
        """create a vector whose elements equal the sixth column of the matrix

        :return: sixth column of the matrix
        :rtype: Vector6D
        """
        return Vector6D(
            self.row1.vz,
            self.row2.vz,
            self.row3.vz,
            self.row4.vz,
            self.row5.vz,
            self.row6.vz,
        )

    def transpose(self) -> "Matrix6D":
        """create a matrix whose rows equal the columns of the calling matrix

        :return: transpose of the calling matrix
        :rtype: Matrix6D
        """
        return Matrix6D(
            self.column_1(),
            self.column_2(),
            self.column_3(),
            self.column_4(),
            self.column_5(),
            self.column_6(),
        )

    def multiply_matrix_6by3(self, mat: Matrix6by3) -> Matrix6by3:
        """create a matrix equal to the product of the calling and argument matrices

        :param mat: matrix to be used in the product
        :type mat: Matrix6by3
        :return: product matrix
        :rtype: Matrix6by3
        """
        return Matrix6by3(
            Vector3D(
                self.row1.dot(mat.column_1()),
                self.row1.dot(mat.column_2()),
                self.row1.dot(mat.column_3()),
            ),
            Vector3D(
                self.row2.dot(mat.column_1()),
                self.row2.dot(mat.column_2()),
                self.row2.dot(mat.column_3()),
            ),
            Vector3D(
                self.row3.dot(mat.column_1()),
                self.row3.dot(mat.column_2()),
                self.row3.dot(mat.column_3()),
            ),
            Vector3D(
                self.row4.dot(mat.column_1()),
                self.row4.dot(mat.column_2()),
                self.row4.dot(mat.column_3()),
            ),
            Vector3D(
                self.row5.dot(mat.column_1()),
                self.row5.dot(mat.column_2()),
                self.row5.dot(mat.column_3()),
            ),
            Vector3D(
                self.row6.dot(mat.column_1()),
                self.row6.dot(mat.column_2()),
                self.row6.dot(mat.column_3()),
            ),
        )

    def multiply_matrix(self, mat: "Matrix6D") -> "Matrix6D":
        """create a matrix equal to the product of the calling and the argument matrices

        :param mat: matrix to be used in the product
        :type mat: Matrix6D
        :return: product matrix
        :rtype: Matrix6D
        """
        return Matrix6D(
            Vector6D(
                self.row1.dot(mat.column_1()),
                self.row1.dot(mat.column_2()),
                self.row1.dot(mat.column_3()),
                self.row1.dot(mat.column_4()),
                self.row1.dot(mat.column_5()),
                self.row1.dot(mat.column_6()),
            ),
            Vector6D(
                self.row2.dot(mat.column_1()),
                self.row2.dot(mat.column_2()),
                self.row2.dot(mat.column_3()),
                self.row2.dot(mat.column_4()),
                self.row2.dot(mat.column_5()),
                self.row2.dot(mat.column_6()),
            ),
            Vector6D(
                self.row3.dot(mat.column_1()),
                self.row3.dot(mat.column_2()),
                self.row3.dot(mat.column_3()),
                self.row3.dot(mat.column_4()),
                self.row3.dot(mat.column_5()),
                self.row3.dot(mat.column_6()),
            ),
            Vector6D(
                self.row4.dot(mat.column_1()),
                self.row4.dot(mat.column_2()),
                self.row4.dot(mat.column_3()),
                self.row4.dot(mat.column_4()),
                self.row4.dot(mat.column_5()),
                self.row4.dot(mat.column_6()),
            ),
            Vector6D(
                self.row5.dot(mat.column_1()),
                self.row5.dot(mat.column_2()),
                self.row5.dot(mat.column_3()),
                self.row5.dot(mat.column_4()),
                self.row5.dot(mat.column_5()),
                self.row5.dot(mat.column_6()),
            ),
            Vector6D(
                self.row6.dot(mat.column_1()),
                self.row6.dot(mat.column_2()),
                self.row6.dot(mat.column_3()),
                self.row6.dot(mat.column_4()),
                self.row6.dot(mat.column_5()),
                self.row6.dot(mat.column_6()),
            ),
        )

    def plus(self, mat: "Matrix6D") -> "Matrix6D":
        """create a matrix whose elements are equal to the sum of the elements in the calling and argument matrices

        :param mat: matrix to be used in the sum
        :type mat: Matrix6D
        :return: sum matrix
        :rtype: Matrix6D
        """
        return Matrix6D(
            self.row1.plus(mat.row1),
            self.row2.plus(mat.row2),
            self.row3.plus(mat.row3),
            self.row4.plus(mat.row4),
            self.row5.plus(mat.row5),
            self.row6.plus(mat.row6),
        )

    def minus(self, mat: "Matrix6D") -> "Matrix6D":
        """create a matrix whose elements are equal to the difference of the elements in the calling and argument matrices

        :param mat: matrix to be used in the difference
        :type mat: Matrix6D
        :return: difference matrix
        :rtype: Matrix6D
        """
        return Matrix6D(
            self.row1.minus(mat.row1),
            self.row2.minus(mat.row2),
            self.row3.minus(mat.row3),
            self.row4.minus(mat.row4),
            self.row5.minus(mat.row5),
            self.row6.minus(mat.row6),
        )



from math import cos, radians, sin

from pyxis.math.functions import Conversions
from pyxis.math.linalg import Vector3D
from pyxis.time import Epoch


class Earth:

    """Class used to store Earth properties base on the WGS84 model"""

    #: G*M given in km^3/s^2
    MU: float = 398600.4418

    #: distance from earth center to surface at the equator in km
    RADIUS: float = 6378.137

    #: value defining the ellipsoid of an oblate earth
    FLATTENING: float = 1 / 298.257223563

    #: inclination of ecliptic relative to earth equator in radians
    OBLIQUITY_OF_ECLIPTIC: float = radians(23.43929111)

    @staticmethod
    def obliquity_of_ecliptic_at_epoch(epoch: Epoch) -> float:
        """calculate the obliquity of ecliptic (epsilon) at a given epoch

        :param epoch: time of interest
        :type epoch: Epoch
        :return: true-of-date epsilon
        :rtype: float
        """
        # Equation 5.39
        t = epoch.julian_centuries_past_j2000()

        a = Conversions.dms_to_radians(0, 0, 46.815)
        b = Conversions.dms_to_radians(0, 0, 0.00059)
        c = Conversions.dms_to_radians(0, 0, 0.001813)

        # Equation 5.42
        return Earth.OBLIQUITY_OF_ECLIPTIC - a * t - b * t * t + c * t * t * t


class Sun:

    """class used to store properties of the sun for a force model"""

    #: RAAN + argument of periapsis in radians
    W_PLUS_W = radians(282.94)

    #: cosine of obliquity of ecliptic
    COS_OBLIQUITY = cos(Earth.OBLIQUITY_OF_ECLIPTIC)

    #: sine of obliquity of ecliptic
    SIN_OBLIQUITY = sin(Earth.OBLIQUITY_OF_ECLIPTIC)

    #: G*M in km^3/s^
    MU = 1.327124400419e11

    #: Estimate of srp
    P = 4.56e-6

    #: Distance to earth in km
    AU = 149597870.691

    @staticmethod
    def get_position(epoch: Epoch) -> Vector3D:
        """calculate the ECI position at a given epoch

        :param epoch: time of calculated position vector
        :type epoch: Epoch
        :return: ECI position in km
        :rtype: Vector3D
        """
        a = Conversions.dms_to_radians(0, 0, 6892)
        b = Conversions.dms_to_radians(0, 0, 72)
        t = epoch.julian_centuries_past_j2000()

        ma = radians(357.5256 + 35999.049 * t)

        sma = sin(ma)
        cma = cos(ma)
        c2ma = cos(2 * ma)

        lam = Sun.W_PLUS_W + ma + a * sma + b * 2 * sma * cma
        r = (149.619 - 2.499 * cma - 0.021 * c2ma) * 1e6

        x = r * cos(lam)

        slam = sin(lam)

        y = r * slam * Sun.COS_OBLIQUITY
        z = r * slam * Sun.SIN_OBLIQUITY

        return Vector3D(x, y, z)


class Moon:
    """class used to store properties of the moon to be used in force modeling"""

    #: G*M in km^3/s^2
    MU = 4902.800305555

    #: distance from center of moon to surface in km
    RADIUS = 1737.4000

    @staticmethod
    def get_position(epoch: Epoch) -> Vector3D:
        """calculate ECI position of moon

        :param epoch: time of calculated position vector
        :type epoch: Epoch
        :return: ECI position in km
        :rtype: Vector3D
        """
        # Equation 3.47
        t = epoch.julian_centuries_past_j2000()
        l0 = radians(218.31617 + 481267.88088 * t - 1.3972 * t)
        l = radians(134.96292 + 477198.86753 * t)
        lp = radians(357.52543 + 35999.04944 * t)
        f = radians(93.27283 + 483202.01873 * t)
        d = radians(297.85027 + 445267.11135 * t)

        # Auxiliary variables to store the angles defined in seconds in equation 3.48
        a0 = Conversions.dms_to_radians(0, 0, 22640.0)
        a1 = Conversions.dms_to_radians(0, 0, 769.0)
        a2 = Conversions.dms_to_radians(0, 0, 4586.0)
        a3 = Conversions.dms_to_radians(0, 0, 2370.0)
        a4 = Conversions.dms_to_radians(0, 0, 668.0)
        a5 = Conversions.dms_to_radians(0, 0, 412.0)
        a6 = Conversions.dms_to_radians(0, 0, 212.0)
        a7 = Conversions.dms_to_radians(0, 0, 206.0)
        a8 = Conversions.dms_to_radians(0, 0, 192.0)
        a9 = Conversions.dms_to_radians(0, 0, 165.0)
        a10 = Conversions.dms_to_radians(0, 0, 148.0)
        a11 = Conversions.dms_to_radians(0, 0, 125.0)
        a12 = Conversions.dms_to_radians(0, 0, 110.0)
        a13 = Conversions.dms_to_radians(0, 0, 55.0)

        # Equation 3.48
        lam = (
            l0
            + a0 * sin(l)
            + a1 * sin(2 * l)
            - a2 * sin(l - 2.0 * d)
            + a3 * sin(2.0 * d)
            - a4 * sin(lp)
            - a5 * sin(2.0 * f)
            - a6 * sin(2.0 * l - 2.0 * d)
            - a7 * sin(l + lp - 2.0 * d)
            + a8 * sin(l + 2.0 * d)
            - a9 * sin(lp - 2.0 * d)
            + a10 * sin(l - lp)
            - a11 * sin(d)
            - a12 * sin(l + lp)
            - a13 * sin(2.0 * f - 2.0 * d)
        )

        # Redefined variables for angles in equation 3.49
        a0 = Conversions.dms_to_radians(0, 0, 18520.0)
        a1 = Conversions.dms_to_radians(0, 0, 412.0)
        a2 = Conversions.dms_to_radians(0, 0, 541.0)
        a3 = Conversions.dms_to_radians(0, 0, 526.0)
        a4 = Conversions.dms_to_radians(0, 0, 44.0)
        a5 = Conversions.dms_to_radians(0, 0, 31.0)
        a6 = Conversions.dms_to_radians(0, 0, 25.0)
        a7 = Conversions.dms_to_radians(0, 0, 23.0)
        a8 = Conversions.dms_to_radians(0, 0, 21.0)
        a9 = Conversions.dms_to_radians(0, 0, 11.0)

        # Equation 3.49
        beta = (
            a0 * sin(f + lam - l0 + a1 * sin(2.0 * f + a2 * sin(lp)))
            - a3 * sin(f - 2.0 * d)
            + a4 * sin(l + f - 2.0 * d)
            - a5 * sin(-l + f - 2.0 * d)
            - a6 * sin(-2.0 * l + f)
            - a7 * sin(lp + f - 2.0 * d)
            + a8 * sin(-l + f)
            + a9 * sin(-lp + f - 2.0 * d)
        )

        # Equation 3.50
        r = (
            385000.0
            - 20905.0 * cos(l)
            - 3699.0 * cos(2.0 * d - l)
            - 2956.0 * cos(2.0 * d)
            - 570.0 * cos(2.0 * l)
            + 246.0 * cos(2.0 * l - 2.0 * d)
            - 205.0 * cos(lp - 2.0 * d)
            - 171.0 * cos(l + 2.0 * d)
            - 152.0 * cos(l + lp - 2.0 * d)
        )

        # Equation 3.51
        x = r * cos(lam) * cos(beta)
        y = r * sin(lam) * cos(beta)
        z = r * sin(beta)

        eps = Earth.OBLIQUITY_OF_ECLIPTIC

        return Vector3D(x, y, z).rotation_about_axis(Vector3D(1, 0, 0), eps)



from math import radians, tan


class OpticalConstraints:

    #: nominal angle at which sensor damage will occur (radians)
    DEFAULT_SUN_HARD = radians(60)

    #: nominal sun angle at which an object can be observed (radians)
    DEFAULT_SUN_SOFT = radians(90)

    #: nominal moon angle at which an object can be observed (radians)
    DEFAULT_MOON_LIMIT = radians(10)

    #: nominal earth angle at which an object can be observed (radians)
    DEFAULT_EARTH_LIMIT = radians(20)

    #: visual magnitude require before an object is considered detected
    DEFAULT_LIMITING_MAG = 15

    #: nominal resolution required for a successful characterization (km)
    DEFAULT_RESOLUTION_LIMIT = 1e-5

    #: nominal ratio of FOV an object can be off center for detection to occur (must be > 0 and <= 1)
    DEFAULT_FRAME_DAMP = 1

    def __init__(self, pixels: float, fov: float) -> None:
        """class to be used for storing limits of an optical observation system

        :param pixels: 1-D full width pixel count assuming a square sensor
        :type pixels: float
        :param fov: half-angle field-of-view in radians
        :type fov: float
        """
        #: minimum sun angle from sensor bore to prevent damage (radians)
        self.sun_hard: float = OpticalConstraints.DEFAULT_SUN_HARD

        #: minimum sun angle from bore to enable detection (radians)
        self.sun_soft: float = OpticalConstraints.DEFAULT_SUN_SOFT

        #: minimum moon angle from bore to enable detection (radians)
        self.moon: float = OpticalConstraints.DEFAULT_MOON_LIMIT

        #: minimum earth angle from bore to enable detection (radians)
        self.earth: float = OpticalConstraints.DEFAULT_EARTH_LIMIT

        #: limiting visual magnitude of the sensor (higher means the sensor will detect dimmer objects)
        self.vismag: float = OpticalConstraints.DEFAULT_LIMITING_MAG

        #: maximum angle an object can be from the center of the sensor (radians)
        self.bore: float = OpticalConstraints.DEFAULT_FRAME_DAMP * fov

        clos: float = OpticalConstraints.DEFAULT_RESOLUTION_LIMIT * 2 * pixels

        #: maximum range an object can be away for successful characterization (km)
        self.characterization: float = clos / tan(fov)



from pyxis.astro.coordinates import HillState
from pyxis.astro.propagators.relative import Hill
from pyxis.estimation.obs import PositionOb
from pyxis.math.constants import SECONDS_IN_DAY
from pyxis.math.linalg import Matrix3by6, Matrix3D, Matrix6by3, Matrix6D, Vector3D, Vector6D
from pyxis.time import Epoch


class RelativeKalman:

    #: Covariance matrix to be used when the filter is initialized (km and km/s)
    DEFAULT_COVARIANCE = Matrix6D(
        Vector6D(0.5, 0, 0, 0, 0, 0),
        Vector6D(0, 0.5, 0, 0, 0, 0),
        Vector6D(0, 0, 0.5, 0, 0, 0),
        Vector6D(0, 0, 0, 5e-5, 0, 0),
        Vector6D(0, 0, 0, 0, 5e-5, 0),
        Vector6D(0, 0, 0, 0, 0, 5e-5),
    )

    #: Processing noise of the system
    DEFAULT_NOISE = Matrix6D(
        Vector6D(1e-16, 0, 0, 0, 0, 0),
        Vector6D(0, 1e-16, 0, 0, 0, 0),
        Vector6D(0, 0, 1e-16, 0, 0, 0),
        Vector6D(0, 0, 0, 1e-16, 0, 0),
        Vector6D(0, 0, 0, 0, 1e-16, 0),
        Vector6D(0, 0, 0, 0, 0, 1e-16),
    )

    #: Observation matrix of a position-only system
    H: Matrix3by6 = Matrix3by6(
        Vector6D(1, 0, 0, 0, 0, 0),
        Vector6D(0, 1, 0, 0, 0, 0),
        Vector6D(0, 0, 1, 0, 0, 0),
    )

    #: Transposed H
    HT: Matrix6by3 = H.transpose()

    #: Matrix with a diagonal of all 1 and off-diagonal of 0
    I: Matrix6D = Matrix6D.identity()

    def __init__(self, epoch: Epoch, propagator: Hill) -> None:
        """class used to act as the navigation filter for a position-only measurement source

        :param epoch: time filter is to be initialized
        :type epoch: Epoch
        :param propagator: estimated Hill state of the spacecraft being observed
        :type propagator: Hill
        """
        #: current time of the filter state
        self.epoch: Epoch = epoch.copy()

        #: propagator used to estimate the relative state of the observed spacecraft
        self.propagator: Hill = propagator

        #: current state of the filter
        self.x00: Vector6D = self.propagator.state.vector.copy()

        #: predicted state of the filter
        self.x10: Vector6D

        #: current covariance of the state
        self.p00: Matrix6D = RelativeKalman.DEFAULT_COVARIANCE

        #: predicted covariance of the state
        self.p10: Matrix6D

        #: processing noise of the filter
        self.q: Matrix6D = RelativeKalman.DEFAULT_NOISE

        #: state transition matrix for the filter state
        self.f: Matrix6D = self.propagator.system_matrix(0)

        #: gain of the filter
        self.k: Matrix6by3

        #: current measurement
        self.z: Vector3D

        #: error in km of the current measurement
        self.range_error: float = 0

        #: measurement uncertainty matrix
        self.r: Matrix3D

    def predict_covariance(self) -> None:
        """calculate the next covariance"""
        self.p10 = self.f.multiply_matrix(self.p00.multiply_matrix(self.f.transpose())).plus(self.q)

    def gain(self) -> None:
        """calculate the gain of the system"""
        normal_measurment = self.z.normalized()
        self.r = Matrix3D(
            Vector3D(
                normal_measurment.x * self.range_error * normal_measurment.x * self.range_error,
                0,
                0,
            ),
            Vector3D(
                0,
                normal_measurment.y * self.range_error * normal_measurment.y * self.range_error,
                0,
            ),
            Vector3D(
                0,
                0,
                normal_measurment.z * self.range_error * normal_measurment.z * self.range_error,
            ),
        )
        hph: Matrix3D = self.H.multiply_matrix_6by3(self.p10.multiply_matrix_6by3(self.HT))
        self.k = self.p10.multiply_matrix_6by3(self.HT.multiply(hph.plus(self.r).inverse()))

    def predict_state(self, dt: float) -> None:
        """calculate the expected future state of the system

        :param dt: number of seconds between current observation and previous observation
        :type dt: float
        """
        self.f = self.propagator.system_matrix(dt)
        self.x10 = self.f.multiply_vector(self.x00)

    def update_state(self) -> None:
        """correct predicted state and store as current state"""
        self.x00 = self.x10.plus(self.k.multiply_vector(self.z.minus(self.H.multiply_vector(self.x10))))
        self.propagator.state = HillState.from_state_vector(self.x00)

    def update_covariance(self) -> None:
        """correct predicted covariance and store as current covariance"""
        m1: Matrix6D = self.I.minus(self.k.multiply_matrix3by6(self.H))
        m2: Matrix6D = self.k.multiply_matrix3by6(self.r.multiply_matrix3by6(self.k.transpose()))
        self.p00 = m1.multiply_matrix(self.p10.multiply_matrix(m1.transpose())).plus(m2)

    def predict(self, dt: float) -> None:
        """calculate predicted state and covariance"""
        self.predict_state(dt)
        self.predict_covariance()

    def update(self) -> None:
        """correct predicted state and covariance"""
        self.gain()
        self.update_state()
        self.update_covariance()

    def process(self, ob: PositionOb) -> None:
        """include a new observation into the current estimation

        :param ob: observation to be added into the system
        :type ob: PositionOb
        """
        dt = (ob.epoch.value - self.epoch.value) * SECONDS_IN_DAY
        self.epoch = ob.epoch.copy()
        self.z = ob.position.copy()
        self.range_error = ob.error
        self.predict(dt)
        self.update()



from math import radians, tan

from pyxis.hardware.constraints import OpticalConstraints


class Camera:

    #: nominal full-length pixel count of a wide field-of-view camera assuming a square sensor
    DEFAULT_WIDE_PIXELS = 3072

    #: nominal full-length pixel count of a narrow field-of-view camera assuming a square sensor
    DEFAULT_NARROW_PIXELS = 3072

    #: nominal half-angle field-of-view in radians for a characterization-type camera
    DEFAULT_NFOV = radians(0.125)

    #: nominal half-angle field-of-view in radians for an observation-type camera
    DEFAULT_WFOV = radians(5)

    def __init__(self, pixels: float, fov: float) -> None:
        """class used to store sensor characteristics for optical observation hardware

        :param pixels: full-length pixel count assuming a square sensor
        :type pixels: float
        :param fov: half-angle field-of-view in radians
        :type fov: float
        """

        #: full-length pixel count assuming a square sensor
        self.pixels: float = pixels

        #: half-angle field-of-view in radians
        self.fov: float = fov

        #: constraints applied to the current camera
        self.limits: OpticalConstraints = OpticalConstraints(self.pixels, self.fov)

        #: scalar used to calculate the resolution at various ranges
        self.resolution_factor = 1 / self.pixels * 0.5

    @classmethod
    def wfov(cls) -> "Camera":
        """method used to instantiate a nominal wide field-of-view camera

        :return: camera containing wfov characteristics
        :rtype: Camera
        """
        return cls(Camera.DEFAULT_WIDE_PIXELS, Camera.DEFAULT_WFOV)

    @classmethod
    def nfov(cls) -> "Camera":
        """method used to instantiate a nominal narrow field-of-view camera

        :return: camera containing nfov characteristics
        :rtype: Camera
        """
        return cls(Camera.DEFAULT_NARROW_PIXELS, Camera.DEFAULT_NFOV)

    def resolution(self, r: float) -> float:
        """calculate the current resolution of the sensor for a given range

        :param r: range in km at which to calculate the resolution capability
        :type r: float
        :return: resolution in km
        :rtype: float
        """
        return tan(self.fov) * r * self.resolution_factor

    def range_error(self, r: float, d: float) -> float:
        """calculate the range estimation error at a given range when observing an object of the input diameter

        :param r: range in km that is being estimated
        :type r: float
        :param d: diameter in km of the object being observed
        :type d: float
        :return: error of the estimation in km
        :rtype: float
        """
        return self.resolution(r) / d * r

    def tracking_minimum(self, body_r: float) -> float:
        """calculate the minimum distance before the object being observed is too large to fit in the frame

        :param body_r: radius of the object being observed
        :type body_r: float
        :return: distance in km at which the fov will be filled
        :rtype: float
        """
        return body_r / tan(self.fov)



from pyxis.math.linalg import Vector3D
from pyxis.time import Epoch


class PositionOb:
    def __init__(self, epoch: Epoch, position: Vector3D, error: float) -> None:
        """class used to store an observation that contains position only

        :param epoch: time of the observation
        :type epoch: Epoch
        :param position: observed position vector in km
        :type position: Vector3D
        :param error: error associated with the magnitude of the observed position vector in km
        :type error: float
        """
        #: time of the observation
        self.epoch: Epoch = epoch.copy()

        #: observed position in km
        self.position: Vector3D = position.copy()

        #: error in km of the observed position magnitude
        self.error: float = error



from math import cos, sin, sqrt

from pyxis.astro.bodies.celestial import Earth
from pyxis.astro.coordinates import HillState
from pyxis.math.linalg import Matrix6D, Vector6D


class Hill:

    #: Nominal time to advance the propagator in seconds when no dt is given
    DEFAULT_STEP_SIZE: float = 600

    def __init__(self, state: HillState, sma: float) -> None:
        """class used to calculate the relative position of a spacecraft using Hill's equations

        :param state: relative state of the satellite to be propagated
        :type state: HillState
        :param sma: semi-major axis of the origin vehicle in km
        :type sma: float
        """
        #: current state of the propagator
        self.state: HillState = state.copy()

        #: semi-major axis of the origin spacecraft
        self.sma: float = sma

        #: orbital rate of the system
        self.n: float = sqrt(Earth.MU / (sma * sma * sma))

        #: step in seconds used to advance the propagator
        self.step_size = Hill.DEFAULT_STEP_SIZE

    def system_matrix(self, t: float) -> Matrix6D:
        """return the system matrix required to advance the initial state by the input t

        :param t: number of seconds between initial state and desired state
        :type t: float
        :return: system matrix used to advance the propagator
        :rtype: Matrix6D
        """
        n = self.n
        n_inv = 1 / n
        sn = sin(n * t)
        cs = cos(n * t)

        # define system matrix of x, y, z, x_dot, y_dot, z_dot equations
        sys_mat = Matrix6D(
            Vector6D(4 - 3 * cs, 0, 0, sn * n_inv, 2 * (1 - cs) * n_inv, 0),
            Vector6D(
                6 * (sn - n * t),
                1,
                0,
                -2 * (1 - cs) * n_inv,
                (4 * sn - 3 * n * t) * n_inv,
                0,
            ),
            Vector6D(0, 0, cs, 0, 0, sn * n_inv),
            Vector6D(3 * n * sn, 0, 0, cs, 2 * sn, 0),
            Vector6D(-6 * n * (1 - cs), 0, 0, -2 * sn, 4 * cs - 3, 0),
            Vector6D(0, 0, -n * sn, 0, 0, cs),
        )

        return sys_mat

    def step_by_seconds(self, t: float) -> None:
        """advance the propagator by a variable time

        :param t: number of seconds to advance the propagator
        :type t: float
        """
        self.state = HillState.from_state_vector(self.system_matrix(t).multiply_vector(self.state.vector))

    def step(self) -> None:
        """advance the propagator by one time step"""
        self.step_by_seconds(self.step_size)



from math import cos, sin

from pyxis.astro.bodies.artificial import Spacecraft
from pyxis.astro.coordinates import AzElRange, ITRFstate, LLAstate
from pyxis.math.linalg import Matrix3D, Vector3D
from pyxis.time import Epoch


class GroundSite:
    def __init__(self, lla: LLAstate) -> None:
        """used for operations that require modeling of a terrestrial location

        :param lla: state that holds the latitude, longitude, and altitude of the ground station
        :type lla: LLAstate
        """
        self.lla: LLAstate = lla.copy()
        self.itrf_position: Vector3D = self.lla.itrf_position()
        slamb: float = sin(self.lla.longitude)
        clamb: float = cos(self.lla.longitude)
        spsi: float = sin(self.lla.latitude)
        cpsi: float = cos(self.lla.latitude)
        self.enz_matrix: Matrix3D = Matrix3D(
            Vector3D(-slamb, clamb, 0.0),
            Vector3D(-spsi * clamb, -spsi * slamb, cpsi),
            Vector3D(cpsi * clamb, cpsi * slamb, spsi),
        )

    @classmethod
    def from_itrf_position(cls, itrf: Vector3D) -> "GroundSite":
        """create a groundsite from cartesian coordinates

        :param itrf: earth-fixed position of the groundsite
        :type itrf: Vector3D
        :return: object stationed at the input earth-fixed vector
        :rtype: GroundSite
        """
        return cls(ITRFstate(Epoch(0), itrf, Vector3D(0, 0, 0)).lla_state())

    def enz_position(self, obj_itrf: Vector3D) -> Vector3D:
        """calculates the east-north-zenith coordinates of the argument position

        :param obj_itrf: itrf position of the object of interest
        :type obj_itrf: Vector3D
        :return: transformation of the argument itrf position to the east-north-zenith frame
        :rtype: Vector3D
        """
        return self.enz_matrix.multiply_vector(obj_itrf.minus(self.itrf_position))

    def angles_and_range(self, target: Spacecraft) -> AzElRange:
        """calculate the topo-centric angles and range to the argument spacecraft

        :param target: spacecraft being observed
        :type target: Spacecraft
        :return: azimuth, elevation, and range to the spacecraft from the ground site
        :rtype: AzElRange
        """
        return AzElRange.from_enz(self.enz_position(target.current_state().itrf_position()))



from math import radians

from pyxis.math.constants import HOURS_IN_DAY, MINUTES_IN_DAY, MINUTES_IN_HOUR, SECONDS_IN_DAY, SECONDS_IN_HOUR


class Conversions:
    """class used to perform various unit conversions

    :return: defined by each static method
    :rtype: defined by each static method
    """

    @staticmethod
    def hms_to_decimal_day(hr: float, m: float, s: float) -> float:
        """calculate the float value of the day given an hour, minute, second representation

        :param hr: hour of day
        :type hr: float
        :param m: minute of day
        :type m: float
        :param s: second of day
        :type s: float
        :return: the time in days
        :rtype: float
        """
        return hr / HOURS_IN_DAY + m / MINUTES_IN_DAY + s / SECONDS_IN_DAY

    @staticmethod
    def dms_to_radians(d: float, m: float, s: float) -> float:
        """calculate an angle in radians that has been defined in degrees, minutes, and seconds

        :param d: degrees in angle
        :type d: float
        :param m: minute of angle
        :type m: float
        :param s: second of angle
        :type s: float
        :return: angle in radians
        :rtype: float
        """
        return radians(d + m / MINUTES_IN_HOUR + s / SECONDS_IN_HOUR)


def sign(num: float) -> float:
    """function to determine if a value is positive or negative

    :param num: expression to be signed
    :type num: float
    :return: 1 if positive -1 if negative 0 if neither
    :rtype: float
    """
    val = 0
    if num > 0:
        val = 1
    elif num < 0:
        val = -1
    return val



from math import floor, radians

from pyxis.math.constants import SECONDS_IN_DAY
from pyxis.math.functions import Conversions


class Epoch:

    #: julian value to act as the MJD start
    MJD_ZERO = 2400000.5

    #: number of days in a julian century
    JULIAN_CENTURY = 36525

    #: julian value for the j2000 epoch
    J2000_JULIAN_DATE = 2451545

    #: offset in days of the universal coordinated time system and the terrestrial dynamic time system
    UTC_TDT_OFFSET = 69.184 / SECONDS_IN_DAY

    def __init__(self, value: float) -> None:
        """class used to represent time

        :param value: time in modified julian days
        :type value: float
        """
        self.value = value

    def copy(self) -> "Epoch":
        """used to create a copy of the current epoch

        :return: a replica of the calling epoch
        :rtype: Epoch
        """
        return Epoch(self.value)

    @classmethod
    def from_gregorian(cls, year: int, month: int, day: int, hour: int, minute: int, sec: float) -> "Epoch":
        """instantiate an epoch from the standard calendar format

        :param year: 4-digit year
        :type year: int
        :param month: 2-digit month
        :type month: int
        :param day: 2-digit day
        :type day: int
        :param hour: 2-digit hour
        :type hour: int
        :param minute: 2-digit minute
        :type minute: int
        :param sec: 2-digit and trailing decimal second
        :type sec: float
        :return: epoch with an mjd value equivalent to the corresponding calendar date
        :rtype: Epoch
        """
        y = year
        m = month
        d = day

        if m <= 2:
            y -= 1
            m += 12

        b = floor(y / 400) - floor(y / 100) + floor(y / 4)
        mjd = 365 * y - 679004 + b + floor(30.6001 * (m + 1)) + d

        return cls(mjd + Conversions.hms_to_decimal_day(hour, minute, sec))

    def julian_value(self) -> float:
        """calculate the julian value of the calling epoch

        :return: full julian date
        :rtype: float
        """
        return self.value + self.MJD_ZERO

    def julian_centuries_past_j2000(self) -> float:
        """calculate the number of julian centuries that have elapsed since the j2000 epoch

        :return: number of julian centuries past the j2000 epoch
        :rtype: float
        """
        return (self.julian_value() - self.J2000_JULIAN_DATE) / self.JULIAN_CENTURY

    def plus_days(self, t: float) -> "Epoch":
        """calculate an epoch that is separated from the calling epoch by t days

        :param t: time delta of the two epochs in days
        :type t: float
        :return: an epoch that is t days away from the calling epoch
        :rtype: Epoch
        """
        return Epoch(self.value + t)

    def greenwich_hour_angle(self) -> float:
        """calculate the greenwich hour angle used to determine sidereal time

        :return: greenwich mean sidereal time in radians
        :rtype: float
        """
        # solve for julian centuries since j2000 using equation 2.7
        utc = Epoch(self.value - Epoch.UTC_TDT_OFFSET)
        dec_day = utc.value % 1
        j0 = utc.julian_value() - dec_day
        j = (j0 - Epoch.J2000_JULIAN_DATE) / Epoch.JULIAN_CENTURY

        # solve for theta0 using equation 2.6
        theta0 = 100.4606184 + 36000.77004 * j + 0.000387933 * j * j

        # solve for gmst using equation 2.8
        total_deg = theta0 + 360.98564724 * dec_day
        return radians(total_deg % 360)



#:
SECONDS_IN_DAY: float = 86400

#:
HOURS_IN_DAY: float = 24

#:
MINUTES_IN_DAY: float = 1440

#:
MINUTES_IN_HOUR: float = 60

#:
SECONDS_IN_HOUR: float = 3600

#:
SECONDS_IN_MINUTE: float = 60

#:
DEGREES_IN_HOUR: float = 15



from math import cos, log10, pi, radians, sin
from random import gauss, uniform
from typing import List

from pyxis.astro.bodies.celestial import Earth
from pyxis.astro.coordinates import GCRFstate, HillState
from pyxis.astro.propagators.inertial import RK4
from pyxis.astro.propagators.relative import Hill
from pyxis.estimation.filtering import RelativeKalman
from pyxis.estimation.obs import PositionOb
from pyxis.hardware.payloads import Camera
from pyxis.math.constants import SECONDS_IN_DAY
from pyxis.math.linalg import Vector3D
from pyxis.time import Epoch


class Spacecraft:

    #: The default body radius of the satellite (km)
    DEFAULT_RADIUS: float = 0.005

    #: The default albedo of the satellite used for optical modeling (unitless)
    DEFAULT_ALBEDO: float = 0.3

    #: Available control methods of the vehicle's attitude (lvlh == z to earth, solar == -z to sun, target == z to rso)
    STEERING_MODES: List[str] = ["lvlh", "solar", "target"]

    #: Default scalar to use with calculating slew times (1/(radians per day))
    DEFAULT_SLEW_SCALE: float = 1 / (radians(0.5) * SECONDS_IN_DAY)

    #: Default tolerance to use for statistical attitude modeling (radians)
    DEFAULT_POINTING_ACCURACY: float = 1e-5

    def __init__(self, state: GCRFstate) -> None:
        """class used to model the behaviors of man-made satellites

        :param state: starting inertial state of the satellite
        :type state: GCRFstate
        """

        #: Used to retain knowledge of the state when the satellite was first created
        self.initial_state: GCRFstate = state.copy()

        #: Used to solve the state of the spacecraft at various times in the orbit
        self.propagator: RK4 = RK4(self.initial_state)

        #: Used for optical modeling of the satellite
        self.albedo: float = Spacecraft.DEFAULT_ALBEDO

        #: Used for various physical modeling methods of the satellite
        self.body_radius: float = Spacecraft.DEFAULT_RADIUS

        #: Payload used for metric observation and close-proximity tracking
        self.wfov: Camera = Camera.wfov()

        #: Payload used for distant tracking and characterization
        self.nfov: Camera = Camera.nfov()

        #: Used for state estimation when target-tracking
        self.filter: RelativeKalman

        #: Used to store the current steering mode of the satellite
        self.steering: str = Spacecraft.STEERING_MODES[0]

        #: The target satellite when the calling spacecraft is in target-tracking mode
        self.tracked_target: Spacecraft

        #: Indicates whether the spacecraft is in a stable attitude mode or in a transitioning slew
        self.slewing: bool = False

        #: Epoch used to indicate when self.slewing can be switched to 'False'
        self.slew_stop: Epoch

        #: Used to calculate slew times
        self.slew_scalar: float = Spacecraft.DEFAULT_SLEW_SCALE

        #: Used to apply noise to attitude vectors
        self.pointing_accuracy = Spacecraft.DEFAULT_POINTING_ACCURACY

        self.update_attitude()

    def sma(self) -> float:
        """calculate the semi-major axis of the calling spacecraft

        :return: the spacecraft's semi-major axis in km
        :rtype: float
        """
        r = self.position().magnitude()
        v = self.velocity().magnitude()
        return 1 / (2 / r - v * v / Earth.MU)

    def acquire(self, seed: "Spacecraft") -> None:
        """initialize the kalman filter and begin tracking the argument satellite

        :param seed: the estimated state of the satellite to be tracked
        :type seed: Spacecraft
        """

        self.filter = RelativeKalman(
            self.current_epoch(),
            Hill(
                HillState.from_gcrf(seed.current_state(), self.current_state()),
                self.sma(),
            ),
        )
        self.track_state(seed)

    def observe_wfov(self, target: "Spacecraft") -> PositionOb:
        """produce a simulated observation from the wfov

        :param target: satellite to be observed
        :type target: Spacecraft
        :return: simulated observation
        :rtype: PositionOb
        """
        # Create hill state of self to target
        tgt = HillState.from_gcrf(target.current_state(), self.current_state())

        # Calculate the range error based on sensor settings
        r = tgt.position.magnitude()
        err = self.wfov.range_error(r, target.body_radius * 2)

        # Apply noise to the range estimate
        ob_r = gauss(r, err)

        # Apply noise to the angular estimate
        ang_err = gauss(0, self.pointing_accuracy)
        ob = tgt.position.normalized().rotation_about_axis(tgt.position.cross(Vector3D(0, 0, 1)), ang_err)
        ob = ob.rotation_about_axis(tgt.position, uniform(0, 2 * pi))

        return PositionOb(self.current_epoch(), ob.scaled(ob_r), err)

    def observe_nfov(self, target: "Spacecraft") -> PositionOb:
        """produce a simulated observation from the nfov

        :param target: satellite to be observed
        :type target: Spacecraft
        :return: simulated observation
        :rtype: PositionOb
        """
        # Create hill state of self to target
        tgt = HillState.from_gcrf(target.current_state(), self.current_state())

        # Calculate the range error based on sensor settings
        r = tgt.position.magnitude()
        err = self.nfov.range_error(r, target.body_radius * 2)

        # Apply noise to the range estimate
        ob_r = gauss(r, err)

        # Apply noise to the angular estimate
        ang_err = gauss(0, self.pointing_accuracy)
        ob = tgt.position.normalized().rotation_about_axis(tgt.position.cross(Vector3D(0, 0, 1)), ang_err)
        ob = ob.rotation_about_axis(tgt.position, uniform(0, 2 * pi))

        return PositionOb(self.current_epoch(), ob.scaled(ob_r), err)

    def process_wfov(self, target: "Spacecraft") -> None:
        """create a simulated observation of the argument spacecraft and feed the ob into the kalman filter

        :param target: satellite to be observed
        :type target: Spacecraft
        """
        ob = self.observe_wfov(target)
        self.filter.process(ob)

    def process_nfov(self, target: "Spacecraft") -> None:
        """create a simulated observation of the argument spacecraft and feed the ob into the kalman filter

        :param target: satellite to be observed
        :type target: Spacecraft
        """
        ob = self.observe_nfov(target)
        self.filter.process(ob)

    def update_attitude(self) -> None:
        """calculate and store the appropriate body axis values depending on steering mode"""
        if self.steering == Spacecraft.STEERING_MODES[0]:

            # Point payload deck at Earth
            self.body_z = self.earth_vector()

            # Align solar panels with orbit normal
            self.body_y = self.position().cross(self.velocity())

            # Complete right-hand rule
            self.body_x = self.body_y.cross(self.body_z)

        elif self.steering == Spacecraft.STEERING_MODES[1]:

            # Point payload deck away from Sun
            self.body_z = self.sun_vector().scaled(-1)

            # Create arbitrary x
            self.body_x = self.body_z.cross(Vector3D(0, 0, 1))

            # Complete right-hand rule
            self.body_y = self.body_z.cross(self.body_x)

        elif self.steering == Spacecraft.STEERING_MODES[2]:

            # Step the tracked spacecraft if epochs are not in sync
            if self.tracked_target.current_epoch().value != self.current_epoch().value:
                self.tracked_target.step_to_epoch(self.current_epoch())

            # Point payload deck at target
            self.body_z = self.target_vector(self.tracked_target)

            # Align solar panels
            self.body_y = self.body_z.cross(self.sun_vector())

            # Complete right-hand rule
            self.body_x = self.body_y.cross(self.body_z)

        if self.slewing:
            if self.current_epoch().value > self.slew_stop.value:
                self.slewing = False

    def track_lvlh(self) -> None:
        """store attitude with payload toward earth and panels along orbit normal"""
        # Include slew duration if not already in lvlh
        if self.steering != Spacecraft.STEERING_MODES[0]:
            self.steering = Spacecraft.STEERING_MODES[0]
            self.slewing = True
            t: float = self.body_z.angle(self.position().scaled(-1)) * self.slew_scalar
            self.slew_stop = self.current_epoch().plus_days(t)

        self.update_attitude()

    def track_sun(self) -> None:
        """store attitude with payload opposite to Sun"""
        # Include slew duration if not already sun-pointing
        if self.steering != Spacecraft.STEERING_MODES[1]:
            self.steering = Spacecraft.STEERING_MODES[1]
            self.slewing = True
            t: float = self.body_z.angle(self.sun_vector().scaled(-1)) * self.slew_scalar
            self.slew_stop = self.current_epoch().plus_days(t)

        self.update_attitude()

    def track_state(self, target: "Spacecraft") -> None:
        """store attitude with payload toward target

        :param target: spacecraft to be tracked
        :type target: Spacecraft
        """
        # Include slew duration if not already target-pointing
        if self.steering != Spacecraft.STEERING_MODES[2]:
            self.steering = Spacecraft.STEERING_MODES[2]
            self.tracked_target = target
            self.slewing = True
            t: float = self.body_z.angle(self.target_vector(target)) * self.slew_scalar
            self.slew_stop = self.current_epoch().plus_days(t)

        self.update_attitude()

    def velocity(self) -> Vector3D:
        """retrieve current ECI velocity vector

        :return: velocity vector as determined by the satellite's propagator
        :rtype: Vector3D
        """
        return self.propagator.state.velocity.copy()

    def detect(self, target: "Spacecraft") -> bool:
        """determine if a satellite can be detected given payload constraints

        :param target: satellite to be detected
        :type target: Spacecraft
        :return: status of detection
        :rtype: bool
        """
        success: bool = True
        if self.sun_angle(target) < self.wfov.limits.sun_soft:
            success = False
        elif self.earth_angle(target) < self.wfov.limits.earth:
            success = False
        elif self.moon_angle(target) < self.wfov.limits.moon:
            success = False
        elif self.visual_magnitude(target) > self.wfov.limits.vismag:
            success = False
        elif self.body_z.angle(self.target_vector(target)) > self.wfov.limits.bore:
            success = False
        return success

    def visual_magnitude(self, target: "Spacecraft") -> float:
        """calculate the visual magnitude of a satellite

        :param target: satellite being observed
        :type target: Spacecraft
        :return: visual magnitude of the observed spacecraft
        :rtype: float
        """
        # Store the target's body radius
        r: float = self.body_radius

        # Calculate the range to the target
        dist: float = self.range(target)

        # Calculate the sun angle
        phi: float = pi - self.sun_angle(target)

        # Calculate the flux and vismag
        fdiff: float = (2 / 3) * self.albedo * r * r / (pi * dist * dist) * ((sin(phi) + (pi - phi) * cos(phi)))
        return -26.74 - 2.5 * log10(fdiff)

    def sun_angle(self, target: "Spacecraft") -> float:
        """calculate the angle between the sun and target using the calling spacecraft as the vertex

        :param target: satellite being observed
        :type target: Spacecraft
        :return: angle between target and sun vector in radians
        :rtype: float
        """
        return self.sun_vector().angle(self.target_vector(target))

    def moon_angle(self, target: "Spacecraft") -> float:
        """calculate the angle between the moon and target using the calling spacecraft as the vertex

        :param target: satellite being observed
        :type target: Spacecraft
        :return: angle between target and moon vector in radians
        :rtype: float
        """
        return self.moon_vector().angle(self.target_vector(target))

    def earth_angle(self, target: "Spacecraft") -> float:
        """calculate the angle between the earth and target using the calling spacecraft as the vertex

        :param target: satellite being observed
        :type target: Spacecraft
        :return: angle between target and earth vector in radians
        :rtype: float
        """
        return self.earth_vector().angle(self.target_vector(target))

    def range(self, target: "Spacecraft") -> float:
        """calculate the distance from the calling spacecraft to the argument spacecraft

        :param target: spacecraft representing the range vector's head
        :type target: Spacecraft
        :return: distance to the argument spacecraft in km
        :rtype: float
        """
        return self.target_vector(target).magnitude()

    def position(self) -> Vector3D:
        """retrieve the spacecraft's current ECI position vector

        :return: spacecraft's current position vector as determined by the propagator
        :rtype: Vector3D
        """
        return self.current_state().position.copy()

    def step(self) -> None:
        """solve the vehicle's position and velocity at the next time step"""
        self.propagator.step()
        self.update_attitude()

    def step_to_epoch(self, epoch: Epoch) -> None:
        """solve the vehicle's position and velocity at the argument epoch

        :param epoch: desired time at which the vehicle's state should be solved
        :type epoch: Epoch
        """
        self.propagator.step_to_epoch(epoch)
        self.update_attitude()

    def sun_vector(self) -> Vector3D:
        """calculate the ECI vector from the vehicle to the sun

        :return: vector originating at self and terminating at the sun
        :rtype: Vector3D
        """
        return self.current_state().sun_vector()

    def moon_vector(self) -> Vector3D:
        """calculate the ECI vector from the vehicle to the moon

        :return: vector originating at self and terminating at the moon
        :rtype: Vector3D
        """
        return self.current_state().moon_vector()

    def earth_vector(self) -> Vector3D:
        """calculate the ECI vector from the vehicle to the earth

        :return: vector originating at self and terminating at the earth
        :rtype: Vector3D
        """
        return self.position().scaled(-1)

    def target_vector(self, target: "Spacecraft") -> Vector3D:
        """calculate the ECI vector from the vehicle to the argument spacecraft

        :param target: spacecraft acting as the vector head
        :type target: Spacecraft
        :return: vector originating at self and terminating at the target's position
        :rtype: Vector3D
        """
        return target.position().minus(self.position())

    def hill_position(self, target: "Spacecraft") -> Vector3D:
        """calculate the hill position vector from self to the argument spacecraft

        :param target: vehicle acting as the relative position vector head
        :type target: Spacecraft
        :return: vector originating at self and terminating at the target's position in the hill frame
        :rtype: Vector3D
        """
        return HillState.from_gcrf(target.current_state(), self.current_state()).position

    def current_state(self) -> "GCRFstate":
        """retrieve the vehicle's current ECI state

        :return: current ECI state as determined by the propagator
        :rtype: GCRFstate
        """
        return self.propagator.state.copy()

    def current_epoch(self) -> Epoch:
        """retrieve the vehicle's current time

        :return: epoch of the vehicle as determined by the propagator
        :rtype: Epoch
        """
        return self.current_state().epoch.copy()
